// SPDX-License-Identifier: GPL-3.0-or-later
#pragma once

#include "platform/platform.h"
#include "parse.h"

extern bool arg_dry_run;

int fd_set_perms(Item *i, int fd, const char *path, const struct stat *st,
                 CreationMode creation);
int fd_set_xattrs(Item *i, int fd, const char *path);
int fd_set_acls(Item *i, int fd, const char *path, const struct stat *st,
                CreationMode creation);

int path_set_perms(Item *i, const char *path, CreationMode creation);
int path_set_xattrs(Item *i, const char *path);
int path_set_acls(Item *i, const char *path, CreationMode creation);
int path_set_attribute(Item *i, const char *path);
