// SPDX-License-Identifier: GPL-3.0-or-later
#pragma once

#include "parse.h"
#include "perms.h"

int create_item(Context *c, Item *i);
int mkdir_parents(const char *path, mode_t mode);
int write_one_file(Item *i, const char *path);
