// SPDX-License-Identifier: GPL-3.0-or-later
#pragma once

#include <sys/types.h>

int name_to_uid_offline(const char *root, const char *user, uid_t *ret);
int name_to_gid_offline(const char *root, const char *group, gid_t *ret);
void offline_passwd_cache_clear(void);
