// SPDX-License-Identifier: GPL-3.0-or-later
#pragma once

#include "parse.h"

int dir_cleanup(Context *c, const char *path, DIR *d, uint64_t self_mtime_usec,
                uint64_t cutoff_usec, dev_t rootdev, bool keep_first_level,
                int maxdepth);

bool unix_socket_alive(const char *path);
int clean_item(Context *c, Item *i);
int remove_item(Context *c, Item *i);
