// SPDX-License-Identifier: GPL-3.0-or-later
#pragma once

#include <glob.h>
#include <stddef.h>

typedef struct {
    char **patterns;
    size_t count;
} pattern_list;

int glob_brace(const char *pattern, glob_t *pglob, int flags,
               int (*errfunc)(const char *epath, int eerrno));

int glob_in_root(const char *pattern, const char *root,
                 int (*cb)(const char *logical_path, void *userdata),
                 void *userdata);
